@@ -44,6 +44,10 @@
 #include "ops.h"
 #include "debug.h"
 
+
+#define spacep(p)   (*(p) == ' ' || *(p) == '\t')
+
+
 
 struct key_queue_item_s
 {
@@ -69,6 +73,10 @@ typedef struct
   /* This points to the last sig in tmp_uid.  */
   gpgme_key_sig_t tmp_keysig;
 
+  /* Helper with a malloced uppercased fingerprint without '!' to be
+   * used to set the subkey_match flag.  */
+  char *requested_subkey;
+
   /* Something new is available.  */
   int key_cond;
   struct key_queue_item_s *key_queue;
@@ -87,6 +95,9 @@ release_op_data (void *hook)
   /* opd->tmp_uid and opd->tmp_keysig are actually part of opd->tmp_key,
      so we do not need to release them here.  */
 
+  if (opd->requested_subkey)
+    free (opd->requested_subkey);
+
   while (key)
     {
       struct key_queue_item_s *next = key->next;
@@ -150,7 +161,9 @@ keylist_status_handler (void *priv, gpgme_status_code_t code, char *args)
       break;
 
     case GPGME_STATUS_FAILURE:
-      opd->failure_code = _gpgme_parse_failure (args);
+      if (!opd->failure_code
+          || gpg_err_code (opd->failure_code) == GPG_ERR_GENERAL)
+        opd->failure_code = _gpgme_parse_failure (args);
       if (opd->failure_code && !strcmp (args, "option-parser")
           && gpg_err_code (opd->failure_code) == GPG_ERR_GENERAL)
         err = gpg_error (GPG_ERR_INV_ENGINE);
@@ -601,7 +614,7 @@ keylist_colon_handler (void *priv, char *line)
   enum
     {
       RT_NONE, RT_SIG, RT_UID, RT_TFS, RT_SUB, RT_PUB, RT_FPR, RT_FP2, RT_GRP,
-      RT_SSB, RT_SEC, RT_CRT, RT_CRS, RT_REV, RT_SPK
+      RT_SSB, RT_SEC, RT_CRT, RT_CRS, RT_REV, RT_SPK, RT_RVK
     }
   rectype = RT_NONE;
 #define NR_FIELDS 20
@@ -667,6 +680,8 @@ keylist_colon_handler (void *priv, char *line)
     rectype = RT_SSB;
   else if (!strcmp (field[0], "spk") && key)
     rectype = RT_SPK;
+  else if (!strcmp (field[0], "rvk") && key)
+    rectype = RT_RVK;
   else
     rectype = RT_NONE;
 
@@ -1122,6 +1137,39 @@ keylist_colon_handler (void *priv, char *line)
 	      keysig->_last_notation = notation;
 	    }
 	}
+      break;
+
+    case RT_RVK:
+      /* Ignore revocation keys without fingerprint */
+      if (fields >= 10 && *field[9])
+        {
+          gpgme_revocation_key_t revkey = NULL;
+
+          err = _gpgme_key_add_rev_key (key, field[9]);
+          if (err)
+            return err;
+
+          revkey = key->_last_revkey;
+          assert (revkey);
+
+          /* Field 4 has the public key algorithm.  */
+          {
+            int i = atoi (field[3]);
+            if (i >= 1 && i < 128)
+              revkey->pubkey_algo = _gpgme_map_pk_algo (i, ctx->protocol);
+          }
+
+          /* Field 11 has the class (eg, 0x40 means sensitive).  */
+          if (fields >= 11 && field[10][0] && field[10][1])
+            {
+              int key_class = _gpgme_hextobyte (field[10]);
+              if (key_class >= 0)
+                revkey->key_class = key_class;
+              if (field[10][2] == 's')
+                revkey->sensitive = 1;
+            }
+        }
+      break;
 
     case RT_NONE:
       /* Unknown record.  */
@@ -1170,6 +1218,64 @@ _gpgme_op_keylist_event_cb (void *data, gpgme_event_io_t type, void *type_data)
 }
 
 
+/* Return true if PATTERN requests an extackt match.  That is a
+ * sequence of at least 40 hexdigits followed by an exclamation mark
+ * and nothing else.  Returns a pointer to the first hexdigit or
+ * NULL.  */
+static const char *
+exact_match_pattern (const char *pattern)
+{
+  const char *s;
+  int hexlen;
+  const char *start;
+
+  for(s = pattern; *s && spacep (s); s++ )  /* Skip spaces.  */
+    ;
+  if (s[0] == '0' && s[1] == 'x')  /* Skip an optional "0x" prefix.  */
+    s += 2;
+  start = s;
+  hexlen = strspn (s, "0123456789abcdefABCDEF");
+  if (hexlen >= 40 && s[hexlen] == '!')
+    {
+      s += hexlen+1;
+      /* Skip trailing spaces and LFs.  */
+      for (; *s && (spacep (s) || *s == '\n' || *s == '\r'); s++ )
+        ;
+      if (!*s) /* Only spaces seen.  */
+        return start;
+    }
+  return NULL;
+}
+
+
+/* Helper for gpgme_op_keylist_start and ..._start_ext.  */
+static gpg_error_t
+maybe_setup_for_requested_subkey (op_data_t opd, const char *pattern)
+{
+  const char *s;
+  char *p;
+
+  if (pattern && (s=exact_match_pattern (pattern)))
+    {
+      free (opd->requested_subkey);
+      p = opd->requested_subkey = strdup (s);
+      if (!p)
+        return gpg_error_from_syserror ();
+      for (; *p; p++ )
+        {
+          if (*p >= 'a' && *p <= 'z')
+            *p &= ~0x20;  /* Upcase it.  */
+          else if (*p == '!') /* Remove the '!' suffix and stop.  */
+            {
+              *p = 0;
+              break;
+            }
+        }
+    }
+  return 0;
+}
+
+
 /* Start a keylist operation within CTX, searching for keys which
    match PATTERN.  If SECRET_ONLY is true, only secret keys are
    returned.  */
@@ -1196,6 +1302,10 @@ gpgme_op_keylist_start (gpgme_ctx_t ctx, const char *pattern, int secret_only)
   if (err)
     return TRACE_ERR (err);
 
+  err = maybe_setup_for_requested_subkey (opd, pattern);
+  if (err)
+    return TRACE_ERR (err);
+
   err = _gpgme_op_import_init_result (ctx);
   if (err)
     return TRACE_ERR (err);
@@ -1240,6 +1350,22 @@ gpgme_op_keylist_ext_start (gpgme_ctx_t ctx, const char *pattern[],
   if (err)
     return TRACE_ERR (err);
 
+  if (pattern && pattern[0])
+    {
+      int i;
+
+      for (i=0; pattern[i]; i++)
+        {
+          err = maybe_setup_for_requested_subkey (opd, pattern[i]);
+          if (err)
+            return TRACE_ERR (err);
+          /* We can only handle one exact subkey request.  Thus we
+           * stop after the first seen.  */
+          if (opd->requested_subkey)
+            break;
+        }
+    }
+
   err = _gpgme_op_import_init_result (ctx);
   if (err)
     return TRACE_ERR (err);
@@ -1341,6 +1467,16 @@ gpgme_op_keylist_next (gpgme_ctx_t ctx, gpgme_key_t *r_key)
   *r_key = queue_item->key;
   free (queue_item);
 
+  if (opd->requested_subkey && (*r_key)->subkeys && (*r_key)->subkeys->fpr)
+    {
+      /* Time to set the mark.  */
+      gpgme_subkey_t subkey = (*r_key)->subkeys;
+
+      for (; subkey; subkey = subkey->next)
+        if (subkey->fpr && !strcmp (subkey->fpr, opd->requested_subkey))
+          subkey->subkey_match = 1;
+    }
+
   TRACE_SUC ("key=%p (%s)", *r_key,
              ((*r_key)->subkeys && (*r_key)->subkeys->fpr) ?
              (*r_key)->subkeys->fpr : "invalid");
